--[[

	BBs Weapon Framework for Teardown | Version 1.0

	Contains helpful stuff to use when making weapons
	Feel free to use in your own mods
	Credit not necessary, most of this work is built off of others' work anyway, primarily elboydo's Thermite Cannon

	(Library by BubblineBee)

]]

--[[

	Handler for shells

	Has an init method that can be used to modify its functionality
	The default keys must be present in the defaultShell dictionary to be usable with the built-in simulation function

]]
ShellHandler = {
	shellNum = 1,
	maxShells = 500,
	defaultShell = {active = false, doGravity = true, velocity = Vec(0, 0, 0), position = Vec(0, 0, 0), lifetime = 5.0, size = 0.0, acceleration = Vec(0, 0, 0)}
}
ShellHandler.__index = ShellHandler

function ShellHandler:new()
	o = {shells = {}}
	setmetatable(o, self)
	return o
end

function ShellHandler:Initialize(config)
	for key, value in pairs(config) do
		self[key] = value
	end
	for i=1, self.maxShells do
		self.shells[i] = deepCopy(self.defaultShell)
	end
end

--Add a shell to the handler and activate it
function ShellHandler:AddShell(shell)
	self.shells[self.shellNum].active = true
	for key, value in pairs(shell) do
		self.shells[self.shellNum][key] = value
	end

	self.shellNum = (self.shellNum % #self.shells) + 1
end

--Simulates the motion of all active shells
--Callbacks should return true if the shell should be deactivated, false otherwise
--Hit callback is passed the following: shell, hitPos, normal, shape, dt, args ((...))
--Sim callback is passed the following: shell, dt, args ((...))
--Time callback is passed the following: shell, dt, args ((...))
--Callbacks can safely be nil
function ShellHandler:SimulateShells(dt, hitCallback, simCallback, timeCallback, ...)
	local args = {...}
	for key, shell in pairs(self.shells) do
		if (shell.active) then
			if (shell.lifetime < 0.0) then
				if (timeCallback) then
					shell.active = timeCallback(shell, dt, unpack(args))
				else
					shell.active = false
				end
			else
				local predictedVelocity = VecAdd(shell.velocity, VecScale(shell.acceleration, dt))
				if (shell.doGravity) then
					predictedVelocity = VecAdd(predictedVelocity, VecScale(Vec(0, -9.8, 0), dt))
				end
				local point2 = VecAdd(shell.position, VecScale(predictedVelocity, dt))

				if (simCallback) then
					shell.active = simCallback(shell, dt, unpack(args))
				end

				if (hitCallback) then
					local hit, dist, normal, shape = QueryRaycast(shell.position, VecNormalize(VecSub(point2, shell.position)), VecLength(VecSub(point2, shell.position)), shell.size)
					if (hit) then
						local hitPos = VecAdd(shell.position, VecScale(VecNormalize(VecSub(point2, shell.position)), dist))
						shell.active = hitCallback(shell, hitPos, normal, shape, dt, unpack(args))
					end
				end

				DrawLine(shell.position, point2)
				shell.position = point2
				shell.velocity = predictedVelocity
				shell.lifetime = shell.lifetime - dt
			end
		end
	end
end

--Perform a recursive copy on a table
function deepCopy(original)
    local original_type = type(original)
    local copy
    if (original_type == 'table') then
        copy = {}
        for original_key, original_value in next, original, nil do
            copy[deepCopy(original_key)] = deepCopy(original_value)
        end
        setmetatable(copy, deepCopy(getmetatable(original)))
    else -- number, string, boolean, etc
        copy = original
    end
    return copy
end